/*
 REACTIVE AUTONOMOUS BLACKHOLE LIST SERVER
 COPYRIGHT (C) 2005 DEEP LOGIC, INC.

This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.

*/

#ifdef HAVE_CONFIG_H
#include <auto-config.h>
#endif

#include <string.h>
#include <stdlib.h>
#include <stdio.h>
#include <math.h>
#include <pthread.h>
#include "btree.h"

static struct btree_node *
btree_node_create (unsigned long key, void *ptr)
{
  struct btree_node *node;

  node = (struct btree_node *) calloc (1, sizeof (struct btree_node));
  if (node == NULL)
    return NULL;

  node->key = key;
  node->ptr = ptr;

  return node;
}

struct btree *
btree_create (void)
{
  struct btree *btree = (struct btree *) malloc (sizeof (struct btree));

  if (btree == NULL)
    return NULL;
  pthread_rwlock_init(&btree->lock, NULL);
  btree->items = 0;
  btree->root = NULL;
  return btree;
}

/* find largest value in tree */
struct btree_node *
btree_first (struct btree *t)
{
  struct btree_node *p;
  if (!t || !t->root)
    return NULL;
  p = t->root;
  while (p->left)
    p = p->left;
  return p;
}

/* find next smallest value in tree */
struct btree_node *
btree_next (struct btree_node *cur)
{
  struct btree_node *p;

  if (cur == NULL)
    return NULL;

  p = cur->right;
  if (p)
  {
    while (p->left) 
      p = p->left;
    return p;
  }

  p = cur->parent;
  if (!p) {
    return NULL;
  }

  while(p->key > cur->key) {
    p = p->parent;
    if (!p) {
      return NULL;
    }
  }

  return p;
}

/* traverse tree in postorder to delete. */
int
btree_destroy (struct btree *btree)
{
  struct btree_node *p, *q;
  int cnt;
  if (btree == NULL)
    return 0;
  cnt = btree->items;
  q = btree_first (btree);
  while (q)
  {
    /* find leftmost leaf node */
    while (q->right)
    {
      q = q->right;
      while (q->left)
        q = q->left;
    }
    /* left and right have been deleted, so safe to delete. */
    do
    {
      p = q;
      q = p->parent;
      free(p->ptr);
      free (p);
      --cnt;
    }
    while (q && q->right == p);
  }
  pthread_rwlock_destroy(&btree->lock);
  free(btree);
  return cnt;
}

struct btree_node *btree_find (struct btree *btree, unsigned long key) {
  struct btree_node *search;

  if (btree == NULL || btree->root == NULL)
    return NULL;

  search = btree->root;

  while(search != NULL && key != search->key) {
    if (key > search->key)
      search = search->left;
    else if (key < search->key)
      search = search->right;
  }

  if (!search)
    return NULL;
  return search;
}

struct btree_node *btree_find_parent (struct btree *btree, unsigned long key) {
  struct btree_node *parent = NULL;
  struct btree_node *search;

  if (btree == NULL || btree->root == NULL)
    return NULL;

  search = btree->root;

  while(search != NULL && key != search->key) {
    parent = search;
    if (key > search->key)
      search = search->left;
    else if (key < search->key)
      search = search->right;
  }

  if (!search)
    return NULL;

  return parent;
}


int
btree_del (struct btree *btree, unsigned long key)
{
  struct btree_node *parent, *del;

  if (btree == NULL)
    return -1;

  del = btree_find(btree, key);
  if (del == NULL) 
    return -1;

  parent = del->parent;
  if (parent == NULL)
    parent = btree->root;

  /* Replace the root node */
  if (del == btree->root) {
    if (del->left != NULL)
      btree->root = btree->root->left;
    else { 
      btree->root = btree->root->right;
      if (btree->root)
        btree->root->parent = NULL;
      free(del);
      btree->items--;
      return 0;
    }
    parent = btree->root;
    if (btree->root)
      btree->root->parent = NULL;

  /* Replace non-root node */
  } else {
    if (del->left != NULL) {
      if (parent->left == del)  
        parent->left = del->left;
      else 
        parent->right = del->left;
      if (del->left)
        del->left->parent = parent;
      parent = del->left;
    } else {
      if (parent->left == del) 
        parent->left = del->right;
      else
        parent->right = del->right;
      if (del->right)
        del->right->parent = parent;
      free(del);
      btree->items--;
      return 0;
    }
  }

  if (del->right == NULL) {
    free(del);
    btree->items--;
    return 0;
  }

  while(parent->right != NULL) 
    parent = parent->right;

  parent->right = del->right;
  if (del->right)
    del->right->parent = parent;
  free(del);
  btree->items--;
  return 0;
}

int
btree_add (struct btree *btree, unsigned long key, void *ptr)
{
  struct btree_node *node;
  struct btree_node *search, *parent;
  int d = 0, left = 1, right = -1;

  if (btree == NULL)
    return -1;

  node = btree_node_create (key, ptr);
  if (node == NULL)
    return -1;

  if (btree->root == NULL)
  {
    btree->root = node;
    btree->items++;
    return 0;
  }

  parent = btree->root;
  search = parent;

  if (key > search->key)
  {
    d = left;
    search = parent->left;
  }
  else
  {
    d = right;
    search = parent->right;
  }

  while (search != NULL)
  {
    parent = search;
    if (key > search->key)
    {
      search = parent->left;
      d = left;
    }
    else
    {
      d = right;
      search = parent->right;
    }
  }

  node->parent = parent;
  if (d == left)
    parent->left = node;
  else
    parent->right = node;

  btree->items++;
  return 0;
}
