/*
 REACTIVE AUTONOMOUS BLACKHOLE LIST SERVER
 COPYRIGHT (C) 2005 DEEP LOGIC, INC.

This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.
                                                                                                                                             
This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.
                                                                                                                                             
You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
                                                                                                                                             
*/

#ifdef HAVE_CONFIG_H
#include <auto-config.h>
#endif

#include <string.h>
#include <stdlib.h>
#include <stdio.h>
#include <pthread.h>
#include "hash.h"

static unsigned long ht_prime_list[ht_num_primes] = {
  53ul,         97ul,         193ul,       389ul,       769ul,
  1543ul,       3079ul,       6151ul,      12289ul,     24593ul,
  49157ul,      98317ul,      196613ul,    393241ul,    786433ul,
  1572869ul,    3145739ul,    6291469ul,   12582917ul,  25165843ul,
  50331653ul,   100663319ul,  201326611ul, 402653189ul, 805306457ul,
  1610612741ul, 3221225473ul, 4294967291ul
};


long ht_hashcode(struct hash *ht, const char *key) {
  unsigned long val = 0, mod;

  if (key == NULL || key[0] == 0 || ht == NULL || ht->size == 0)
    return 0;

  for (; *key; ++key) 
    val = 5 * val + *key;
  mod = val % ht->size;
  return mod;
}

struct ht_node *ht_node_create(const char *key) {
  char *node_key;
  struct ht_node* node;

  node = (struct ht_node *) malloc(sizeof(struct ht_node));
  if (node == NULL) 
    return NULL;

  if ((node_key = strdup(key)) == NULL) {
    free(node);
    return NULL;
  }

  node->key = node_key;
  node->ptr = NULL;
  node->next = NULL;
  return(node);
}

struct hash *ht_create(long size) {
  unsigned long i = 0;
  struct hash *ht = (struct hash *) malloc(sizeof(struct hash));

  if (ht == NULL)
    return NULL;

  pthread_rwlock_init(&ht->lock, NULL);
  while (ht_prime_list[i] < size) 
    i++;
  ht->size  = ht_prime_list[i];
  ht->items = 0;
  ht->tbl   = (struct ht_node **) calloc(ht->size, sizeof(struct ht_node *));
  if (ht->tbl == NULL) {
    free(ht);
    return NULL;
  }
  return(ht);
}

int ht_destroy(struct hash *ht) {
  struct ht_node *node, *next;
  struct ht_c c;

  if (ht == NULL) 
    return -1;

  node = c_ht_first(ht, &c);
  while(node != NULL) {
    next = c_ht_next(ht, &c);
    free(node->ptr);
    ht_delete(ht, node->key);
    node = next;
  }

  free(ht->tbl);
 
  pthread_rwlock_destroy(&ht->lock);
  free(ht);
  ht = (struct hash *) NULL;
  return 0;
}

struct ht_node *ht_find(struct hash *ht, const char *key) {
  long hash_code;
  struct ht_node *node;
  char *k = (char *) key;

  if (ht == NULL || key[0] == 0) 
    return NULL;

  hash_code = ht_hashcode(ht, k);
  node = ht->tbl[hash_code];
                                                                                                                                             
  while (node) {
    if (node->key != NULL && !strcmp(key, node->key)) 
      return(node);
    node = node->next;
  }
  return NULL;
}

void *ht_ptr(struct hash *ht, const char *key) {
  long hash_code;
  struct ht_node *node;
  char *k = (char *) key;

  if (ht == NULL || key[0] == 0) 
    return NULL;

  hash_code = ht_hashcode(ht, k);
  node = ht->tbl[hash_code];

  while (node) {
    if (node->key != NULL && !strcmp(key, node->key)) {
      if (node->ptr == NULL) 
        return NULL; 
      else 
        return(node->ptr);
    }
    node = node->next;
  }
  return NULL;
}

int ht_set(struct hash *ht, const char *key, void *val) {
  long hash_code;
  struct ht_node *parent;
  struct ht_node *node;
  struct ht_node *new_node = NULL;

  if (ht == NULL) 
    return -1;

  hash_code	= ht_hashcode(ht, key);
  parent	= NULL;
  node		= ht->tbl[hash_code];

  while (node != NULL) {
    if (!strcmp(key, node->key)) {
      new_node = node;
      node = NULL;
    }

    if (new_node == NULL) {
      parent = node;
      node = node->next;
    }
  }

  if (new_node != NULL) {
    new_node->ptr = val;
    return 0;
  } 

  /* Create a new node */
  new_node = ht_node_create(key);
  new_node->ptr = val;

  ht->items++;

  /* Insert */
  if (parent) {  /* Existing parent */
    parent->next = new_node;
    return 0;
  }

  /* No existing parent; add directly to hash table */
  ht->tbl[hash_code] = new_node;

  return 0;
}

int ht_delete(struct hash *ht, const char *key) {
  long hash_code;
  struct ht_node *node;
  struct ht_node *del_node = NULL;
  struct ht_node *parent = NULL;
  int ret = 0;

  if (ht == NULL) 
    return -1;

  hash_code = ht_hashcode(ht, key);
  node = ht->tbl[hash_code];

  while (node) {
    if (!strcmp(key, node->key)) {
      del_node = node;
      node = NULL; 
    } else {
      parent = node;
      node = node->next;
    }
  }

  if (del_node == NULL) 
    return -3;
    
  if (parent) 
    parent->next = del_node->next;
  else 
    ht->tbl[hash_code] = del_node->next;

  free(del_node->key);
  free(del_node);
  ht->items--;

  return ret;
}

struct ht_node *c_ht_first(struct hash *ht, struct ht_c *c) {
  c->iter_index = 0;
  c->iter_next = (struct ht_node *)NULL;
  return(c_ht_next(ht, c));
}

struct ht_node *c_ht_next(struct hash *ht, struct ht_c *c) {
  long index;
  struct ht_node *node = c->iter_next;

  if (ht == NULL) return NULL;

  if (node) {
    c->iter_next = node->next;
    return (node);
  } else {
    while (c->iter_index < (ht->size-1)) {
      index = c->iter_index++;
      if (ht->tbl[index]) {
        c->iter_next = ht->tbl[index]->next;
        return(ht->tbl[index]);
      }
    }
  }

  return((struct ht_node *)NULL);
}

