/*
 REACTIVE AUTONOMOUS BLACKHOLE LIST SERVER
 COPYRIGHT (C) 2005 DEEP LOGIC, INC.

This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.

*/

#ifdef HAVE_CONFIG_H
#include <auto-config.h>
#endif

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <errno.h>
#include <ctype.h>

#ifdef TIME_WITH_SYS_TIME
#   include <sys/time.h>
#   include <time.h>
#else
#   ifdef HAVE_SYS_TIME_H
#       include <sys/time.h>
#   else
#       include <time.h>
#   endif
#endif

#include <sys/types.h>
#include <sys/resource.h>
#include <sys/socket.h>
#include <netdb.h>
#include <netinet/in.h>
#include <netinet/tcp.h>
#include <arpa/inet.h>
#include <unistd.h>
#include <fcntl.h>
#include <signal.h>
#include <pthread.h>

#include "rabl_server.h"
#include "error.h"
#include "btree.h"
#include "hash.h"
#include "language.h"
#include "md5.h"

int
main (int argc, char *argv[])
{
  int i;

  config_file = NULL;

  printf("RABL Server v%s\n", VERSION); 

  for(i=0;i<argc;i++) {
    if (!strcmp(argv[i], "-c") && i+1<argc) {
      config_file = strdup(argv[i+1]);
    } else if (!strcmp(argv[i], "-u") && i+1<argc) {
      setuid(atoi(argv[i+1]));
    }
  }

  if (config_file == NULL) {
    ERROR(SYNTAX);
    return EXIT_FAILURE;
  }

  read_config(0);

  if (config == NULL) {
    ERROR(SYNTAX);
    return EXIT_FAILURE;
  }

  signal(SIGHUP, &reload_config);
  signal(SIGTERM, &dieout);
  signal(SIGINT, &dieout);

  daemon_listen();

  destroy_config();
  pthread_exit(0);
  return EXIT_SUCCESS;
}

void dieout(int sig) {
  fprintf(stderr, "terminating on signal %d\n", sig);
  destroy_config();
  free(config_file);
  exit(EXIT_SUCCESS);
}

void read_config(int sig) {
  FILE *file;
  char buffer[1024];
  char *token;
  struct rabl_auth *auth;

  if (sig) {
    LOG(LOG_WARNING, "resetting RABL on signal %d", sig);
  }

  /* Reload */
  destroy_config();

  config = calloc(1, sizeof(struct rabl_server_config));
  if (config == NULL)
    return; 

  pthread_rwlock_init(&config->lock, NULL);
  config->log = NULL;
  config->doscmd = NULL;
  config->zone_template = NULL;
  config->zone_output = NULL;
  config->auth = btree_create();
  config->peers = btree_create();
  config->rabl_ipv4 = btree_create();
  config->rabl_ipv6 = ht_create(24593ul);
  config->conn_count = ht_create(1543ul);
  config->connections = btree_create();
 
  /* Configuration Defaults */
  config->port = 1039;
  config->cleanup = 600;
  config->minspread = 10;
  config->queue = 32;
  config->life = 86400;

  if (config->auth == NULL       || 
      config->peers == NULL      ||
      config->rabl_ipv4 == NULL   || 
      config->rabl_ipv6 == NULL   || 
      config->conn_count == NULL) 
  {
    destroy_config();
    config = NULL;
    return;
  }
 
  file = fopen(config_file, "r");
  if (file == NULL) {
    LOG(LOG_WARNING, "unable to open configuration file '%s': %s", config_file, strerror(errno)); 
    config = NULL;
    return;
  }

  while(fgets(buffer, sizeof(buffer), file)!=NULL) {
    if (buffer[0] == '#') 
      continue;

    token = strtok(buffer, "\t ");
    if (token == NULL)
      continue;

     if (!strcasecmp(token, "USER")) {
       char *uid, *shared_secret, *acl, *privileges;
       auth = calloc(1, sizeof(struct rabl_auth));
       if (auth == NULL) {
         LOG(LOG_CRIT, ERROR_MEM_ALLOC);
         continue;
       }

       uid = strtok(NULL, "\t ");
       shared_secret = strtok(NULL, "\t ");
       acl = strtok(NULL, "\t ");
       privileges = strtok(NULL, "\t \n");

       if (   uid  == NULL ||
    shared_secret  == NULL ||
               acl == NULL ||
        privileges == NULL) 
         continue;

       auth->uid		= strtol(uid, NULL, 0);
       auth->shared_secret	= strdup(shared_secret);
       auth->acl		= strdup(acl);

       if (!strcasecmp(privileges, "RW"))
         auth->privileges = AUTH_RDWR;
       else if (!strcasecmp(privileges, "SERVER"))
         auth->privileges = AUTH_SERVER;
       else
         auth->privileges = AUTH_NONE;

       btree_add(config->auth, auth->uid, auth);

     } else if (!strcasecmp(token, "PORT")) {
       char *port;

       port = strtok(NULL, "\t \n");
       if (port != NULL)
       config->port = atoi(port);

     } else if (!strcasecmp(token, "QUEUE")) {
       char *queue;
                                                                                                                                   
       queue = strtok(NULL, "\t \n");
       if (queue != NULL)
       config->queue = atoi(queue);

     } else if (!strcasecmp(token, "CLEANUP")) {
       char *cleanup;

       cleanup = strtok(NULL, "\t \n");
       if (cleanup != NULL)
       config->cleanup = atoi(cleanup);
     } else if (!strcasecmp(token, "MINSPREAD")) {
       char *minspread;

       minspread = strtok(NULL, "\t \n");
       if (minspread != NULL)
         config->minspread = atoi(minspread);
     } else if (!strcasecmp(token, "LIFE")) {
       char *life;

       life = strtok(NULL, "\t \n");
       if (life != NULL)
         config->life = strtol(life, NULL, 0);
     } else if (!strcasecmp(token, "LOGFILE")) {
       char *log;

       log = strtok(NULL, "\t \n");
       if (log != NULL) 
         config->log = strdup(log);
     } else if (!strcasecmp(token, "DOSCMD")) {
       char *cmd;
       cmd = strtok(NULL, "\"");
       if (cmd != NULL)
         config->doscmd = strdup(cmd);
     } else if (!strcasecmp(token, "ZONETEMPLATE")) {
       char *filename;
       filename = strtok(NULL, "\t \n");
       if (filename != NULL)
         config->zone_template = strdup(filename);
     } else if (!strcasecmp(token, "ZONEOUTPUT")) {
       char *filename;
       filename = strtok(NULL, "\t \n");
       if (filename != NULL)
         config->zone_output = strdup(filename);
     } 
  }
  fclose(file);

  return;
}

int destroy_config(void) {

  if (config != NULL) {
    struct btree_node *node_btree;
    struct rabl_auth *auth;
    struct rabl_address *address;
    struct ht_node *node_ht;
    struct ht_c c_ht;

    pthread_rwlock_wrlock(&config->lock);

    node_btree = btree_first(config->auth);
    while(node_btree != NULL) {
      auth = node_btree->ptr; 
      if (auth != NULL) {
        free(auth->shared_secret);
        free(auth->acl);
      }
      node_btree = btree_next(node_btree);
    }

    node_btree = btree_first(config->rabl_ipv4);
    while(node_btree != NULL) {
      address = node_btree->ptr;
      btree_destroy(address->reporting_networks); 
      node_btree = btree_next(node_btree);
    }

    node_ht = c_ht_first(config->rabl_ipv6, &c_ht);
    while(node_btree != NULL) {
      address = node_btree->ptr;
      btree_destroy(address->reporting_networks);
      node_ht = c_ht_next(config->rabl_ipv6, &c_ht);
    }
    
    free(config->log);
    free(config->doscmd);
    free(config->zone_template);
    free(config->zone_output);
    ht_destroy(config->conn_count);
    btree_destroy(config->auth);
    btree_destroy(config->peers);
    btree_destroy(config->rabl_ipv4);
    ht_destroy(config->rabl_ipv6);
    btree_destroy(config->connections);
    free(config);
    config = NULL;
  }

  return 0;
}

int destroy_user_config(void) {

  if (config != NULL) {
    struct btree_node *node_btree;
    struct rabl_auth *auth;

    node_btree = btree_first(config->auth);
    while(node_btree != NULL) {
      auth = node_btree->ptr;
      if (auth != NULL) {
        free(auth->shared_secret);
        free(node_btree->ptr);
      }
      node_btree->ptr = NULL;
      node_btree = btree_next(node_btree);
    }

    free(config->log);
    btree_destroy(config->auth);
    btree_destroy(config->peers);
  }
                                                                                
  return 0;
}

int daemon_listen(void) {
  struct sockaddr_in local_addr;
  struct sockaddr_in remote_addr;
  struct timeval tv;
  connection *conn = NULL;
  time_t local = time(NULL);
  time_t last_object_cleanup = local;
  int recv_len;
  int fdmax;
  int listener;
  int yes = 1;
  int i;
  pthread_attr_t attr;

  FD_ZERO(&master);
  FD_ZERO(&read_fds);

  pthread_attr_init(&attr);
  pthread_attr_setdetachstate(&attr, PTHREAD_CREATE_DETACHED);

#ifdef DEBUG
  printf("creating listening socket on port %d...\n", config->port);
#endif

  listener = socket(AF_INET, SOCK_STREAM, 0);
  if (listener == -1) {
    ERROR("could not create socket: %s", strerror(errno));
    return(EFAILURE);
  }

  if (setsockopt(listener,SOL_SOCKET,SO_REUSEADDR,&yes,sizeof(int)) == -1) {
    ERROR("could not set sockopt SO_REUSEADDR: %s", strerror(errno));
    return(EFAILURE);
  }

  memset(&local_addr, 0, sizeof(struct sockaddr_in));
  local_addr.sin_family = AF_INET;
  local_addr.sin_port = htons(config->port);
  local_addr.sin_addr.s_addr = INADDR_ANY;

#ifdef DEBUG
  printf("binding to :%d...\n", config->port);
#endif

  if (bind(listener, (struct sockaddr *)&local_addr, sizeof(struct sockaddr)) == -1) {
    ERROR("could not bind to :%d: %s", config->port, strerror(errno));
    return(EFAILURE);
  }

  if (listen(listener, config->queue) == -1) {
    ERROR("could not set queue length to %d: %s", config->queue, strerror(errno));
    return(EFAILURE);
  }

#ifdef DEBUG
  fprintf(stderr, "listening on port %d\n", config->port);
#endif

  FD_SET(listener, &master);
  fdmax = listener;
                                                                                                                                   
  conn = malloc(sizeof(connection));
  if (conn == NULL) {
    close(listener);
    LOG(LOG_CRIT, ERROR_MEM_ALLOC);
    return EUNKNOWN;
  }

  conn->sockfd = listener;
  conn->auth = NULL;
  conn->packet_len = 0;
  conn->packet_type = 0;
  conn->invalid_requests = 0;
  conn->recv_len = 0;
  conn->last_receive = local;
  conn->last_full_receive = local;
  conn->data[0] = 0;
  memset(&conn->remote_addr, 0, sizeof(struct sockaddr_in));

  pthread_rwlock_wrlock(&config->connections->lock);
  btree_add(config->connections, listener, (void *) conn);
  pthread_rwlock_unlock(&config->connections->lock);

  for(;;) {
    time_t local = time(NULL);

    read_fds = master;

    tv.tv_sec = 0;
    tv.tv_usec = 50;

    if (select(fdmax+1, &read_fds, NULL, NULL, &tv)>0) {

      /* Process read-ready connections */

      for(i=0;i<=fdmax;i++) {
        if (FD_ISSET(i, &read_fds)) {

          /* ACCEPT NEW CONNECTIONS */
          /* ====================== */

          if (i == listener) {
            int newfd;
            int addrlen = sizeof(remote_addr);
            if ((newfd = accept(listener, (struct sockaddr *)&remote_addr, &addrlen)) == -1) {
              LOG(LOG_WARNING, "unable to accept new connection: %s", strerror(errno));
            } else {
#ifdef DEBUG
              printf("connection id %d from %s.\n", newfd, inet_ntoa(remote_addr.sin_addr));
#endif
              if (check_incoming(&remote_addr)) {
                LOG(LOG_WARNING, "getting flooded with connections from %s", inet_ntoa(remote_addr.sin_addr));
                close(newfd);
                FD_CLR(newfd, &master);
                if (config->doscmd != NULL) {
                  char cmd[256];
                  snprintf(cmd, sizeof(cmd), config->doscmd, inet_ntoa(remote_addr.sin_addr));
                  system(cmd);
                }
                continue;
              }

              fcntl(newfd, F_SETFL, O_NONBLOCK);

              setsockopt(newfd,SOL_SOCKET,TCP_NODELAY,&yes,sizeof(int));

              FD_SET(newfd, &master);
              if (newfd>fdmax) 
                fdmax = newfd;

              conn = malloc(sizeof(connection));
              if (conn == NULL) {
                LOG(LOG_CRIT, ERROR_MEM_ALLOC);
                FD_CLR(newfd, &master);
                free_incoming(&remote_addr);
                close(newfd);
                continue;
              } else { 
                conn->sockfd = newfd;
                conn->auth = NULL;
                conn->packet_len = 0;
                conn->packet_type = 0;
                conn->invalid_requests = 0;
                conn->recv_len = 0;
                conn->last_receive = local;
                conn->last_full_receive = local;
                conn->data[0] = 0;
                memcpy(&conn->remote_addr, &remote_addr, sizeof(remote_addr));

                pthread_rwlock_wrlock(&config->connections->lock);
                btree_add(config->connections, conn->sockfd, (void *) conn);
                pthread_rwlock_unlock(&config->connections->lock);
                send(conn->sockfd, PROTOCOL_VERSION, 1, 0);
              }
            }

          /* PROCESS EXISTING CONNECTIONS */
          /* ============================ */

          } else { /* if (i == listener) .. [else] */
            struct btree_node *node_btree;
            node_btree = btree_find(config->connections, i);

            if (node_btree == NULL || node_btree->ptr == NULL) {
              FD_CLR(i, &master);
              free_incoming(&conn->remote_addr);
              close(i);
              conn = NULL;
              continue;
            }

            conn = node_btree->ptr;

            /* READ THE HEADER */
            /* =============== */

            if (conn->packet_type == 0) {
              recv_len = recv(i, conn->data+conn->recv_len, sizeof(conn->packet_type)-conn->recv_len, 0);
              if (recv_len == 0) {
                FD_CLR(i, &master);
                end_connection(conn);
                conn = NULL;
                continue;
              } else {
                conn->recv_len += recv_len;
                conn->last_receive = local;
                if (conn->recv_len == sizeof(conn->packet_type)) {
                  memcpy(&conn->packet_type, conn->data, sizeof(conn->packet_type)); 
                  if (conn->packet_type == PT_IPV4)
                    conn->packet_len = sizeof(struct rabl_data_ipv4);
                  else if (conn->packet_type == PT_IPV6)
                    conn->packet_len = sizeof(struct rabl_data_ipv6);
                  else if (conn->packet_type == PT_AUTH)
                    conn->packet_len = sizeof(struct rabl_data_auth);
                  else if (conn->packet_type == PT_SIPV4 && conn->auth != NULL && conn->auth->privileges == AUTH_SERVER)
                    conn->packet_len = sizeof(struct rabl_data_sipv4);
                  else if (conn->packet_type == PT_SIPV6 && conn->auth != NULL && conn->auth->privileges == AUTH_SERVER)
                    conn->packet_len = sizeof(struct rabl_data_sipv6);
                  else {
                    FD_CLR(i, &master);
                    end_connection(conn);
                    conn = NULL;
                    continue;
                  }
                  conn->data[0] = 0;
                  conn->recv_len = 0;
                }
              }

            /* READ THE PACKET */
            /* =============== */

            } else if (conn->packet_len > conn->recv_len) {
              recv_len = recv(i, conn->data+conn->recv_len, conn->packet_len-conn->recv_len, 0); 
              conn->recv_len+=recv_len;
              if (recv_len == 0) {
                FD_CLR(i, &master);
                end_connection(conn);
                conn = NULL;
                continue;
              } else {
                conn->last_receive = local;
              }
            }

            /* PROCESS THE PACKET */
            /* ================== */

            if (conn->packet_type && conn->packet_len == conn->recv_len) {
              struct thread_process *process;

              conn->last_full_receive = local;
              FD_CLR(i, &master);

              process = malloc(sizeof(struct thread_process));
              if (process == NULL) {
                LOG(LOG_CRIT, ERROR_MEM_ALLOC);
                end_connection(conn);
                conn = NULL;
                continue;
              }

              process->conn = conn;
              if (pthread_create(&conn->thread, &attr, process_connection, (void *) process)) {
                LOG(LOG_CRIT, "thread creation failed");
                end_connection(conn);
                conn = NULL;
                continue;
              }

              continue;
            }

          } /* if (i == listener) else */

        } else { /* if FD_SET else */
          struct btree_node *node_btree;

          /* TIME OUT STALE CONNECTIONS */
          /* ========================== */

          if (i != listener && FD_ISSET(i, &master)) {

            node_btree = btree_find(config->connections, i);
            if (node_btree == NULL || node_btree->ptr == NULL) {
              FD_CLR(i, &master);
              end_connection(conn);
              conn = NULL;
              continue;
            }

            conn = node_btree->ptr;
  
            if (local-conn->last_full_receive>CONNECTION_TIMEOUT) {
              FD_CLR(i, &master);
              end_connection(conn);
              conn = NULL;
              continue;
            }
 
          } /* if (i != listener) */

        } /* if FD_SET else */

      } /* for(i.. */
  
    }  /* if (select)... */

    /* run periodic cleanups */
    if (local-last_object_cleanup>config->cleanup) {
      pthread_t thread;
      last_object_cleanup = local;
      if (pthread_create(&thread, &attr, object_cleanup, (void *) NULL)) {
        LOG(LOG_CRIT, "thread creation failed");
      }
    }

  } /* for(;;) */

  return 0;
}

void *object_cleanup(void *v) {
  struct btree_node *node_btree, *node_network, *node_next, *node_next2;
  struct ht_node *node_ht, *node_hnext;
  struct ht_c c_ht;
  struct rabl_address *address;
  struct rabl_network *network;
  int active;
  time_t local = time(NULL);
  FILE *in = NULL, *out = NULL;

#ifdef DEBUG
  printf("in object_cleanup\n");
#endif

  if (pthread_rwlock_rdlock(&config->lock))
    goto LOCK;
  if (pthread_rwlock_wrlock(&config->rabl_ipv4->lock))
    goto LOCK2;

  /* Write zone header */
  if (config->zone_template && config->zone_output) {
#ifdef DEBUG
    printf("opening %s -> %s\n", config->zone_template, config->zone_output);
#endif
    in = fopen(config->zone_template, "r");
    if (!in)
    {
      fprintf(stderr, "unable to open %s for reading: %s\n", 
              config->zone_template, strerror(errno));
    } else {
      out = fopen(config->zone_output, "w");
      if (!out)
        fprintf(stderr, "unable to open %s for writing: %s\n", 
                config->zone_output, strerror(errno));
    }

    if (in && out) {
      char buf[1024];
      while(fgets(buf, sizeof(buf), in)!=NULL) {
        if (!strstr(buf, "@serial@")) {
          fprintf(out, "%s", buf);
        } else {
          char *x = strstr(buf, "@serial@");
          x[0] = 0;
          fprintf(out, "%s%ld%s", buf, time(NULL), x+8);
        }
      }
      fclose(in);
    } else {
      if (in) fclose(in);
      if (out) fclose(out);
      in = NULL;
      out = NULL;
    }
  }

  node_btree = btree_first(config->rabl_ipv4);
  while(node_btree != NULL) {
    address = node_btree->ptr;

    node_next2 = btree_next(node_btree);
    if (address != NULL && address->reporting_networks != NULL) {
      active = 0;
      node_network = btree_first(address->reporting_networks);
      while(node_network != NULL) {
        node_next = btree_next(node_network);
        network = node_network->ptr;

        if (network == NULL) {
          btree_del(address->reporting_networks, node_network->key);
}
        if (local - network->last_hit > config->life) {
          free(network);
          btree_del(address->reporting_networks, node_network->key);
        } else {
          active++; 
        }
        
        node_network = node_next;
      }
      if (!active) {
        btree_destroy(address->reporting_networks);
        btree_del(config->rabl_ipv4, node_btree->key);
        free(address);
      } else if (out) {
        if (address->reporting_networks->items >= config->minspread) {
          unsigned char a[4];
          memcpy(a, &node_btree->key, 4);
          fprintf(out, "%d.%d.%d.%d\t\tIN\tA\t127.0.0.2\n", a[3], a[2], a[1], 
            a[0]);
        }
      }
    }
    node_btree = node_next2;
  }

  if (out) fclose(out);
  pthread_rwlock_unlock(&config->rabl_ipv4->lock);

  if (pthread_rwlock_wrlock(&config->rabl_ipv6->lock))
    goto LOCK2;
  node_ht = c_ht_first(config->rabl_ipv6, &c_ht);
  while(node_ht != NULL) {
    node_hnext = c_ht_next(config->rabl_ipv6, &c_ht);

    address = node_ht->ptr;
    if (address != NULL) {
      active = 0;
      node_network = btree_first(address->reporting_networks);
      while(node_network != NULL) {

        node_next = btree_next(node_network);
        network = node_network->ptr;
        if (network == NULL)
          btree_del(address->reporting_networks, node_network->key);
        if (local - network->last_hit > config->life) {
          free(network);
          btree_del(address->reporting_networks, node_network->key);
        } else {
          active++;
        }

        node_network = node_next;
      }
      if (!active) {
        btree_destroy(address->reporting_networks);
        btree_del(config->rabl_ipv4, node_btree->key);
        free(address);
      }
    }
    node_ht = node_hnext;
  }
  pthread_rwlock_unlock(&config->rabl_ipv6->lock);
  pthread_rwlock_unlock(&config->lock);

  return 0;

LOCK2:
  pthread_rwlock_unlock(&config->lock);

LOCK:
  fprintf(stderr, "object_cleanup() unable to obtain rwlock.  aborting cleanup.");

  return 0;
}

int end_connection(connection *conn) {
  free_incoming(&conn->remote_addr);
  pthread_rwlock_wrlock(&config->connections->lock);
  btree_del(config->connections, (long) conn->sockfd);
  pthread_rwlock_unlock(&config->connections->lock);
  close(conn->sockfd);
  free(conn);
  return 0;
}

void * process_connection(void *t) {
  struct thread_process *thread = t;
  int r = EINVAL;

  pthread_rwlock_rdlock(&config->lock);
  pthread_rwlock_rdlock(&config->connections->lock);

  switch (thread->conn->packet_type) {
    case PT_IPV4:	r =  process_ipv4_packet(thread->conn);
			break;
    case PT_IPV6:	r = process_ipv6_packet(thread->conn);
			break;
    case PT_AUTH:	r = process_auth_packet(thread->conn);
			break;
    case PT_SIPV4:	r = process_sipv4_packet(thread->conn);
			break;
    case PT_SIPV6:	r = process_sipv6_packet(thread->conn);
			break;
  }
  pthread_rwlock_unlock(&config->lock);
  pthread_rwlock_unlock(&config->connections->lock);

  if (r) {
    end_connection(thread->conn);
  } else {
    thread->conn->packet_type = 0;
    thread->conn->packet_len = 0;
    thread->conn->recv_len = 0;
    thread->conn->data[0] = 0;
    FD_SET(thread->conn->sockfd, &master);
  }

  free(t);
  pthread_exit(0);
  return NULL;
}

int process_ipv4_packet(connection *conn) {
  struct rabl_data_ipv4 packet;

  memcpy(&packet, conn->data, sizeof(struct rabl_data_ipv4));

  if ((packet.type == DATA_POST_SPAM) && conn->auth == NULL) {
    send(conn->sockfd, "E", 1, 0);
    return EINVAL;
  }

  if (packet.type == DATA_QUERY) {
    struct btree_node *node_btree;
    struct rabl_address *address;
    int blacklisted = 0;

    pthread_rwlock_rdlock(&config->rabl_ipv4->lock);
    node_btree = btree_find(config->rabl_ipv4, packet.address);
    if (node_btree != NULL) {
      address = (struct rabl_address *) node_btree->ptr;
      if (address->reporting_networks->items >= config->minspread)
        blacklisted = 1;
    }

    if (blacklisted)
      send(conn->sockfd, "B", 1, 0);
    else
      send(conn->sockfd, "N", 1, 0);

    pthread_rwlock_unlock(&config->rabl_ipv4->lock);

    return 0;
  }

  if (packet.type == DATA_POST_SPAM) {
    struct btree_node *node_btree;
    struct rabl_address *address;
    struct rabl_network *network;
    int add_spread = 0;

    pthread_rwlock_wrlock(&config->rabl_ipv4->lock);

    node_btree = btree_find(config->rabl_ipv4, packet.address);

    if (node_btree == NULL) {
      address = calloc(1, sizeof(struct rabl_address));
      if (address == NULL) {
        goto MEM_FAIL;
      }
      address->reporting_networks = btree_create();
      if (address->reporting_networks == NULL) {
        free(address);
        goto MEM_FAIL;
      } 
      btree_add(config->rabl_ipv4, packet.address, address);
    } else {
      address = node_btree->ptr;
    }

    node_btree = btree_find(address->reporting_networks, conn->auth->uid);
    if (node_btree == NULL) {
      network = calloc(1, sizeof(struct rabl_network));
      if (network == NULL) 
        goto MEM_FAIL;
      add_spread = 1;
      btree_add(address->reporting_networks, conn->auth->uid, network);
    } else {
      network = node_btree->ptr;
    }

    network->spam_hits++;
    network->last_hit = time(NULL);
    if (config->log != NULL) {
       FILE *file;
       file = fopen(config->log, "a");
       if (file != NULL) {
         char x[4];
         memcpy(&x, &packet.address, 4);
         fprintf(file, "address %d.%d.%d.%d reported from uid %ld\n", 
           (unsigned char) x[0], (unsigned char) x[1], (unsigned char) x[2], 
           (unsigned char) x[3], conn->auth->uid);
         if (add_spread && address->reporting_networks->items == config->minspread) {
           fprintf(file, "address %d.%d.%d.%d now blacklisted\n",
             (unsigned char) x[0], (unsigned char) x[1], (unsigned char) x[2],
             (unsigned char) x[3]);
         }
         fclose(file);
       }
    }

    pthread_rwlock_unlock(&config->rabl_ipv4->lock);
    send(conn->sockfd, "A", 1, 0);	/* Accepted */
  }

  return 0;

MEM_FAIL:
  pthread_rwlock_unlock(&config->rabl_ipv4->lock);
  LOG(LOG_CRIT, ERROR_MEM_ALLOC);
  send(conn->sockfd, "E", 1, 0);
  return EUNKNOWN;
}

int process_ipv6_packet(connection *conn) {
  struct rabl_data_ipv6 packet;

  memcpy(&packet, conn->data, sizeof(struct rabl_data_ipv4));

  if ((packet.type == DATA_POST_SPAM) && conn->auth == NULL) {
    send(conn->sockfd, "E", 1, 0);
    return EINVAL;
  }

  if (packet.type == DATA_QUERY) {
    struct rabl_address *address;
    int blacklisted = 0;

    pthread_rwlock_rdlock(&config->rabl_ipv6->lock);
    address = ht_ptr(config->rabl_ipv6, packet.address);
    if (address != NULL && address->reporting_networks->items >= config->minspread)
      blacklisted = 1;
                                                                                                                   
    if (blacklisted)
      send(conn->sockfd, "B", 1, 0);
    else
      send(conn->sockfd, "N", 1, 0);
    pthread_rwlock_unlock(&config->rabl_ipv6->lock);
    return 0;
  }

  if (packet.type == DATA_POST_SPAM) {
    int add_spread = 0;
    struct btree_node *node_btree;
    struct rabl_address *address;
    struct rabl_network *network;
 
    pthread_rwlock_wrlock(&config->rabl_ipv6->lock);
    address = ht_ptr(config->rabl_ipv6, packet.address);
    if (address == NULL) {
      address = calloc(1, sizeof(struct rabl_address));
      if (address == NULL) {
        goto MEM_FAIL;
      }
      address->reporting_networks = btree_create();
      if (address->reporting_networks == NULL) {
        free(address);
        goto MEM_FAIL;
      }
      ht_set(config->rabl_ipv6, packet.address, address);
    } 

    node_btree = btree_find(address->reporting_networks, conn->auth->uid);
    if (node_btree == NULL) {
      network = calloc(1, sizeof(struct rabl_network));
      if (network == NULL)
        goto MEM_FAIL;
      add_spread = 1;
      btree_add(address->reporting_networks, conn->auth->uid, network);
    } else {
      network = node_btree->ptr;
    }

    network->spam_hits++;
    network->last_hit = time(NULL);
    if (config->log != NULL) {
       FILE *file;
       file = fopen(config->log, "a");
       if (file != NULL) {
         fprintf(file, "address %s reported from uid %ld\n",
           packet.address, conn->auth->uid);
         if (add_spread && address->reporting_networks->items == config->minspread) {
           fprintf(file, "address %s now blacklisted\n",
             packet.address);
         }
         fclose(file);
       }
    }

    pthread_rwlock_unlock(&config->rabl_ipv6->lock);
    send(conn->sockfd, "A", 1, 0);      /* Accepted */
  }

  return 0;
                                                                                                                   
MEM_FAIL:
  pthread_rwlock_unlock(&config->rabl_ipv6->lock);
  LOG(LOG_CRIT, ERROR_MEM_ALLOC);
  send(conn->sockfd, "E", 1, 0);
  return EUNKNOWN;
}

int process_auth_packet(connection *conn) {
  char auth_data[16], response[16], md5[17];
  struct rabl_auth *auth = NULL;
  struct rabl_data_auth packet;
  struct btree_node *node_btree;
  int i,valid = 0,aclok = 0;
  char *acl;

  /* Generate a random 16-byte string, receive 16-byte md5 response */
  for(i=0;i<16;i++) 
    auth_data[i] = (rand() % 254) + 1;
 
  send(conn->sockfd, auth_data, 16, 0);

  if (receive_sock(conn->sockfd, &response, 16)) {
    return EFAILURE;
  }

  pthread_rwlock_rdlock(&config->auth->lock);

  memcpy(&packet, conn->data, sizeof(struct rabl_data_auth));
  node_btree = btree_find(config->auth, packet.uid); 
  if (node_btree != NULL)
    auth = node_btree->ptr;

  if (auth != NULL) {
    MD5_CTX MTX;

    MD5Init (&MTX);
    MD5Update (&MTX, auth_data, 16);
    MD5Update (&MTX, auth->shared_secret, strlen(auth->shared_secret));
    MD5Final (md5, &MTX);

    if (!memcmp(response, md5, 16)) 
      valid = 1;

    /* Check ACL 
       Formats:

         *		Everything
         a.b.c.		Match first 3 octets
         a.b.		Match first 2 octets
         a.b.c.d	Match all octets

         a.b.c.d:a.b.c.d	Multiple ACLs

    */

    if (valid && auth->acl != NULL) {
      char address[64];
      snprintf(address, sizeof(address), "%s", inet_ntoa(conn->remote_addr.sin_addr));
      acl = strtok(auth->acl, ":");

      while(acl != NULL && acl[0] != 0) {
        if (!strcmp(acl, "*")) 
          aclok = 1;
        else if (acl[strlen(acl)-1] != '.' && !strcmp(address, acl))
          aclok = 1;
        else if (acl[strlen(acl)-1] == '.' && !strncmp(address, acl, strlen(acl)))
          aclok = 1;
        acl = strtok(NULL, ":");
      }

      if (!aclok)
        valid = 0;
    }

    if (auth->privileges == AUTH_NONE)
      valid = 0;
  }

  if (valid)
    send(conn->sockfd, "A", 1, 0); /* Accepted */
  else
    send(conn->sockfd, "E", 1, 0); /* Error */

  conn->auth = auth;
  pthread_rwlock_unlock(&config->auth->lock);
  return (valid) ? 0 : EFAILURE;
}

int receive_sock(int sockfd, void *buff, int len) {
  int recv_len = 0, i;
  time_t start = time(NULL);
  struct timeval tv;

  while(recv_len<len && start - time(NULL)<CONNECTION_TIMEOUT) {
    i = recv(sockfd, buff+recv_len, len-recv_len, 0);
    if (i == 0)
      return EFAILURE;
    if (i>0)
      recv_len += i;
    if (i<0) {
      tv.tv_sec = 0;
      tv.tv_usec = 100;

      select(0, NULL, NULL, NULL, &tv);
    }
  }

  if (recv_len == len) 
    return 0;

  return EFAILURE;
}


void reload_config(int sig) {
  FILE *file;
  char buffer[1024];
  char *token;
  struct rabl_auth *auth;
  struct btree_node *node_btree;

  signal(SIGHUP, SIG_IGN);

  if (sig) {
    LOG(LOG_WARNING, "reloading on signal %d", sig);
  }

  pthread_rwlock_wrlock(&config->lock);

  node_btree = btree_first(config->connections);
  while(node_btree != NULL) {
    connection *conn = node_btree->ptr;
    if (conn->thread) {
      pthread_kill(conn->thread, SIGTERM);
      FD_CLR(conn->sockfd, &master);
      end_connection(conn);
    }
    node_btree = btree_next(node_btree);
  }

  destroy_user_config();

  config->log = NULL;
  config->auth = btree_create();
  config->peers = btree_create();
 
  /* Configuration Defaults */
  config->port = 1039;
  config->cleanup = 600;
  config->minspread = 10;
  config->queue = 32;
  config->life = 86400;

  if (config->auth == NULL || config->peers == NULL) {
    destroy_config();
    fprintf(stderr, ERROR_MEM_ALLOC);
    exit(EXIT_FAILURE);
  }
 
  file = fopen(config_file, "r");
  if (file == NULL) {
    LOG(LOG_WARNING, "unable to open configuration file '%s': %s", config_file, strerror(errno)); 
    destroy_config();
    exit(EXIT_FAILURE);
  }

  while(fgets(buffer, sizeof(buffer), file)!=NULL) {
    if (buffer[0] == '#') 
      continue;

    token = strtok(buffer, "\t ");
    if (token == NULL)
      continue;

     if (!strcmp(token, "USER")) {
       char *uid, *shared_secret, *acl, *privileges;
       auth = calloc(1, sizeof(struct rabl_auth));
       if (auth == NULL) {
         LOG(LOG_CRIT, ERROR_MEM_ALLOC);
         continue;
       }

       uid = strtok(NULL, "\t ");
       shared_secret = strtok(NULL, "\t ");
       acl = strtok(NULL, "\t ");
       privileges = strtok(NULL, "\t \n");

       if (   uid  == NULL ||
    shared_secret  == NULL ||
               acl == NULL ||
        privileges == NULL) 
         continue;

       auth->uid		= strtol(uid, NULL, 0);
       auth->shared_secret	= strdup(shared_secret);
       auth->acl		= strdup(acl);

       if (!strcmp(privileges, "RW"))
         auth->privileges = AUTH_RDWR;
       else if (!strcmp(privileges, "SERVER"))
         auth->privileges = AUTH_SERVER;
       else
         auth->privileges = AUTH_NONE;

       btree_add(config->auth, auth->uid, auth);

     } else if (!strcmp(token, "PORT")) {
       char *port;

       port = strtok(NULL, "\t \n");
       if (port != NULL)
       config->port = atoi(port);

     } else if (!strcmp(token, "QUEUE")) {
       char *queue;
                                                                                                                                   
       queue = strtok(NULL, "\t \n");
       if (queue != NULL)
       config->queue = atoi(queue);

     } else if (!strcmp(token, "CLEANUP")) {
       char *cleanup;

       cleanup = strtok(NULL, "\t \n");
       if (cleanup != NULL)
       config->cleanup = atoi(cleanup);
     } else if (!strcmp(token, "MINSPREAD")) {
       char *minspread;

       minspread = strtok(NULL, "\t \n");
       if (minspread != NULL)
         config->minspread = atoi(minspread);
     } else if (!strcmp(token, "LIFE")) {
       char *life;

       life = strtok(NULL, "\t \n");
       if (life != NULL)
         config->life = strtol(life, NULL, 0);
     } else if (!strcmp(token, "LOGFILE")) {
       char *log;

       log = strtok(NULL, "\t \n");
       if (log != NULL) 
         config->log = strdup(log);
     } else if (!strcmp(token, "DOSCMD")) {
       char *cmd;
       cmd = strtok(NULL, "\"");
       if (cmd != NULL)
         config->doscmd = strdup(cmd);
     }

  }
  fclose(file);

  pthread_rwlock_unlock(&config->lock);

  signal(SIGHUP, &reload_config);

  return;
}

int check_incoming(struct sockaddr_in *remote_addr) {
  char address[64];
  struct ht_node *node_ht;
  struct unique_address *u; 

  snprintf(address, sizeof(address), inet_ntoa(remote_addr->sin_addr));
  node_ht = ht_find(config->conn_count, address);
  if (node_ht == NULL) {
    u = calloc(1, sizeof(struct unique_address));
    if (u == NULL) {
      LOG(LOG_CRIT, ERROR_MEM_ALLOC);
      return 0;
    }
    u->count = 0;
    ht_set(config->conn_count, address, (void *) u);
  } else {
    u = node_ht->ptr;
  }

  if (u->count >= MAX_CONN) {
#ifdef DEBUG
    printf("check_incoming() FAIL; connection count for %s: %d\n", address, u->count);
#endif
    return EFAILURE;
  }

  u->count++;

#ifdef DEBUG
  printf("check_incoming() OK; connection count for %s: %d\n", address, u->count);
#endif 

  return 0;
}

int free_incoming(struct sockaddr_in *remote_addr) {
  char address[64];
  struct ht_node *node_ht;
  struct unique_address *u;

  snprintf(address, sizeof(address), inet_ntoa(remote_addr->sin_addr));
  node_ht = ht_find(config->conn_count, address);
  if (node_ht == NULL) {
    return 0;
  }

  u = node_ht->ptr;
  if (!u) {
    ht_delete(config->conn_count, address);
    return 0;
  }

  if (u->count>0)
    u->count--;

  if (u->count == 0) {
    free(node_ht->ptr);
    ht_delete(config->conn_count, address);
  }

#ifdef DEBUG
  printf("free_incoming(): connection count for %s: %d\n", address, u->count);
#endif

  return 0;
}

int process_sipv4_packet(connection *conn) {
  struct rabl_data_sipv4 packet;

  memcpy(&packet, conn->data, sizeof(struct rabl_data_sipv4));

  if ((packet.type == DATA_POST_SPAM) && conn->auth == NULL) {
    send(conn->sockfd, "E", 1, 0);
    return EINVAL;
  }

  if (packet.type == DATA_POST_SPAM) {
    struct btree_node *node_btree;
    struct rabl_address *address;
    struct rabl_network *network;
    int add_spread = 0;

    pthread_rwlock_wrlock(&config->rabl_ipv4->lock);
    node_btree = btree_find(config->rabl_ipv4, packet.address);
    if (node_btree == NULL) {
      address = calloc(1, sizeof(struct rabl_address));
      if (address == NULL) {
        goto MEM_FAIL;
      }
      address->reporting_networks = btree_create();
      if (address->reporting_networks == NULL) {
        free(address);
        goto MEM_FAIL;
      } 
      btree_add(config->rabl_ipv4, packet.address, address);
    } else {
      address = node_btree->ptr;
    }

    node_btree = btree_find(address->reporting_networks, packet.uid);
    if (node_btree == NULL) {
      network = calloc(1, sizeof(struct rabl_network));
      if (network == NULL) 
        goto MEM_FAIL;
      add_spread = 1;
      btree_add(address->reporting_networks, packet.uid, network);
    } else {
      network = node_btree->ptr;
    }

    network->spam_hits++;
    network->last_hit = time(NULL);
    if (config->log != NULL) {
       FILE *file;
       file = fopen(config->log, "a");
       if (file != NULL) {
         char x[4];
         memcpy(&x, &packet.address, 4);
         fprintf(file, "address %d.%d.%d.%d reported from uid %ld via server uid %ld\n", 
           (unsigned char) x[0], (unsigned char) x[1], (unsigned char) x[2], 
           (unsigned char) x[3], (long) packet.uid, conn->auth->uid);
         if (add_spread && address->reporting_networks->items == config->minspread) {
           fprintf(file, "address %d.%d.%d.%d now blacklisted\n",
             (unsigned char) x[0], (unsigned char) x[1], (unsigned char) x[2],
             (unsigned char) x[3]);
         }
         fclose(file);
       }
    }

    pthread_rwlock_unlock(&config->rabl_ipv4->lock);
    send(conn->sockfd, "A", 1, 0);	/* Accepted */
  }

  return 0;

MEM_FAIL:
  pthread_rwlock_unlock(&config->rabl_ipv4->lock);
  LOG(LOG_CRIT, ERROR_MEM_ALLOC);
  send(conn->sockfd, "E", 1, 0);
  return EUNKNOWN;
}

int process_sipv6_packet(connection *conn) {
  struct rabl_data_sipv6 packet;

  memcpy(&packet, conn->data, sizeof(struct rabl_data_ipv4));

  if ((packet.type == DATA_POST_SPAM) && conn->auth == NULL) {
    send(conn->sockfd, "E", 1, 0);
    return EINVAL;
  }

  if (packet.type == DATA_POST_SPAM) {
    int add_spread = 0;
    struct btree_node *node_btree;
    struct rabl_address *address;
    struct rabl_network *network;

    pthread_rwlock_wrlock(&config->rabl_ipv6->lock);
    address = ht_ptr(config->rabl_ipv6, packet.address);
    if (address == NULL) {
      address = calloc(1, sizeof(struct rabl_address));
      if (address == NULL) {
        goto MEM_FAIL;
      }
      address->reporting_networks = btree_create();
      if (address->reporting_networks == NULL) {
        free(address);
        goto MEM_FAIL;
      }
      ht_set(config->rabl_ipv6, packet.address, address);
    } 

    node_btree = btree_find(address->reporting_networks, packet.uid);
    if (node_btree == NULL) {
      network = calloc(1, sizeof(struct rabl_network));
      if (network == NULL)
        goto MEM_FAIL;
      add_spread = 1;
      btree_add(address->reporting_networks, packet.uid, network);
    } else {
      network = node_btree->ptr;
    }

    network->spam_hits++;
    network->last_hit = time(NULL);
    if (config->log != NULL) {
       FILE *file;
       file = fopen(config->log, "a");
       if (file != NULL) {
         fprintf(file, "address %s reported from uid %ld via server uid %ld\n",
           packet.address, (long) packet.uid, conn->auth->uid);
         if (add_spread && address->reporting_networks->items == config->minspread) {
           fprintf(file, "address %s now blacklisted\n",
             packet.address);
         }
         fclose(file);
       }
    }

    pthread_rwlock_unlock(&config->rabl_ipv6->lock);
    send(conn->sockfd, "A", 1, 0);      /* Accepted */
  }

  return 0;
                                                                                                                   
MEM_FAIL:
  pthread_rwlock_unlock(&config->rabl_ipv6->lock);
  LOG(LOG_CRIT, ERROR_MEM_ALLOC);
  send(conn->sockfd, "E", 1, 0);
  return EUNKNOWN;
}

